# -*- coding: utf-8 -*-
"""
Handles instrument specific info for the HARPS spectrograph

Mostly reading data from the header
"""
import logging
import os.path

import numpy as np
from astropy import units as q
from astropy.io import fits
from astropy.time import Time
from dateutil import parser

from .common import Instrument, getter, observation_date_to_night

logger = logging.getLogger(__name__)


class JWST_NIRISS(Instrument):
    def add_header_info(self, header, mode, **kwargs):
        """read data from header and add it as REDUCE keyword back to the header"""
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)
        info = self.load_info()

        # TODO: this references some files, I dont know where they should be
        header["e_gain"] = 1.61
        header["e_readn"] = 18.32
        header["e_dark"] = 0.0257

        # total exposure time
        header["exptime"] = header.get("TFRAME", 0)

        return header

    def split_observation(self, fname, mode):
        hdu = fits.open(fname)
        dirname = os.path.dirname(fname)
        fname = os.path.basename(fname)

        header = hdu[0].header
        if "mjd-obs" not in header:
            if len(header["DATE-OBS"]) <= 10:
                time = header["DATE-OBS"] + "T" + header["TIME-OBS"]
            else:
                time = header["DATE-OBS"]
            header["MJD-OBS"] = Time(time).mjd

        header2 = fits.Header()
        header2["EXTNAME"] = "SCI"
        shape = hdu["SCI"].data.shape

        nframes = shape[0]
        ngroups = shape[1]

        if nframes == 1 and ngroups == 1:
            return [os.path.join(dirname, fname)]

        data = hdu["SCI"].data.reshape((-1, *shape[-2:]))
        bias = data[0]
        primary = fits.PrimaryHDU(header=header)
        files = []
        os.makedirs(os.path.join(dirname, "split"), exist_ok=True)
        for i in range(1, nframes * ngroups):
            this = data[i] - bias
            bias = data[i]
            fname_this = os.path.join(dirname, "split", f"pyreduce_{i}_{fname}")

            header["MJD-OBS"] += header["TFRAME"] * q.s.to(q.day)
            secondary = fits.ImageHDU(data=data, header=header2)
            hdu_this = fits.HDUList([primary, secondary])
            hdu_this.writeto(fname_this, overwrite=True)
            files.append(fname_this)
        return files

    def sort_files(self, input_dir, target, night, mode, allow_calibration_only=False):
        files = super().sort_files(
            input_dir,
            target,
            night,
            mode,
            allow_calibration_only=allow_calibration_only,
        )
        for i, (k, file) in enumerate(files):
            files_split = []
            for f in file["science"]:
                files_split += self.split_observation(f, mode)
            files[i][1]["science"] = files_split
        return files

    def get_wavecal_filename(self, header, mode, **kwargs):
        """Get the filename of the wavelength calibration config file"""
        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_2D.npz".format(instrument="harps", mode=mode)
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname



# -*- coding: utf-8 -*-
"""
with our powers combined we increase snr
"""
import logging
from os.path import dirname, join

import matplotlib.pyplot as plt
import numpy as np
from spectres import spectres
from tqdm import tqdm

from .. import echelle

logger = logging.getLogger(__name__)


def combine(files, output, plot=None):
    # Create a Wavelength grid that will be used for all spectra
    # Based on the one in the "first" fits file
    e = echelle.read(files[0], continuum_normalization=False)
    nord, ncol = e.spec.shape

    # Prepare some empty arrays for storage of all the data
    # TODO what if this becomes to large to handle?
    waves = np.zeros((len(files), nord, ncol))
    specs = np.zeros((len(files), nord, ncol))
    sigms = np.zeros((len(files), nord, ncol))
    conts = np.zeros((len(files), nord, ncol))

    mask = np.full(len(files), True)

    # Load all the data from all files
    # And resample the spectrum and the continuum onto the shared grid
    for k, file in tqdm(enumerate(files), desc="File", total=len(files)):
        try:
            e = echelle.read(file, continuum_normalization=False)
            specs[k] = np.ma.filled(e.spec, 0)
            waves[k] = np.ma.getdata(e.wave)
            sigms[k] = np.ma.filled(e.sig, 1)
            conts[k] = np.ma.filled(e.cont, 0)
        except ValueError as ex:
            logger.warning("Error in loading file %s. %s", file, ex)
            mask[k] = False

    waves = waves[mask]
    specs = specs[mask]
    sigms = sigms[mask]
    conts = conts[mask]

    # wmin, wmax = waves.min(axis=(0, 2)), waves.max(axis=(0, 2))
    # wnew = np.geomspace(wmin, wmax, ncol, endpoint=True).T
    # TODO something weird happens when changing the wavelength grid, that also depends on the wavelength
    # Maybe points in the grid are interpreted differently, i.e. in the rebinning they are the center of the
    # bin, but later on they are the edges?
    wnew = np.copy(waves[0])

    for k in tqdm(range(specs.shape[0]), desc="File"):
        for i in tqdm(range(nord), desc="Order", leave=False):
            conts[k, i], _ = spectres(wnew[i], waves[k, i], conts[k, i], sigms[k, i])
            specs[k, i], sigms[k, i] = spectres(
                wnew[i], waves[k, i], specs[k, i], sigms[k, i]
            )

    # These are just for plotting
    if plot:
        sold = np.copy(specs[:, plot])
        cold = np.copy(conts[:, plot])

    # Median and MAD
    # We use the MAD over the whole range though, since we need some data points to properly evaluate it
    arr = specs / conts
    mean = np.nanmedian(arr, axis=0)
    std = np.nanmedian(np.abs(arr - mean), axis=[0, 2])[:, None]
    vmin, vmax = mean - 5 * std, mean + 5 * std

    # Disregard all values outside of 5 * MAD
    where = (arr < vmin) | (arr > vmax)
    specs[where] = 0
    conts[where] = 1
    sigms[where] = np.nan
    weights = 1 / sigms
    weights[np.isposinf(weights)] = np.sqrt(2)  # TODO Why
    weights[where] = 0

    w2 = np.sum(weights, axis=0) == 0
    weights[:, w2] = 1

    # Take the average of the spectrum and the continuum
    snew = np.average(specs, weights=weights, axis=0)
    cnew = np.average(conts, weights=weights, axis=0)
    snew = np.nan_to_num(snew, copy=False)
    # unew = 1 / np.sqrt(np.nansum((conts/sigms)**2, axis=0))

    # This is the uncertainty from the scatter
    unew = np.sqrt(np.nansum(weights * (arr - snew) ** 2, axis=0) / len(files))
    unew[unew == 0] = np.nansum(sigms, axis=0)[unew == 0]

    snew /= cnew
    cnew = np.ones_like(snew)

    if plot:
        for i in range(sold.shape[0]):
            plt.plot(wnew[plot], sold[i] / cold[i])
        plt.plot(wnew[plot], snew[plot], "--")
        plt.fill_between(wnew[plot], vmin[plot], vmax[plot], alpha=0.5)
        plt.show()

    e.spec = snew
    e.sig = unew
    e.cont = cnew
    e.mask = (snew == 0) | (cnew == 0)
    del e["columns"]

    e.header["barycorr"] = 0.0
    e.save(output)

    logger.info("Created combined file: %s", output)



# -*- coding: utf-8 -*-
"""
Interface for all instrument specific information
The actual info is contained in the instruments/{name}.py modules/classes, which are all subclasses of "common"
"""

import importlib

from .common import Instrument


def load_instrument(instrument) -> Instrument:
    """Load an python instrument module

    Parameters
    ----------
    instrument : str
        name of the instrument

    Returns
    -------
    instrument : Instrument
        Instance of the {instrument} class
    """

    # TODO: Loading arbitrary modules is probably bad style
    # from instruments import uves, harps
    # instruments = {"uves": uves.UVES, "harps": harps.HARPS}
    # instrument = instruments[instrument.lower()]
    # instrument = instrument()
    if instrument is None:
        instrument = "common"

    fname = ".instruments.%s" % instrument.lower()
    lib = importlib.import_module(fname, package="pyreduce")
    instrument = getattr(lib, instrument.upper())
    instrument = instrument()

    return instrument


def get_instrument_info(instrument):
    """Load instrument specific information

    Parameters
    ----------
    instrument : str
        Name of the instrument

    Returns
    -------
    dict{str:obj}
        Dictionary with information
    """

    instrument = load_instrument(instrument)
    return instrument.info


def sort_files(input_dir, target, night, instrument, mode, **kwargs):
    """Sort a list of files into different categories and discard files that are not used

    Parameters
    ----------
    input_dir : str
        directory containing all files (with tags for target, night, and instrument)
    target : str
        observation target name, as found in the files
    night : str
        observation night of interest, as found in the files
    instrument : str
        instrument name
    mode : str
        instrument mode, if applicable (e.g. red/blue for HARPS)

    Returns
    -------
    biaslist : list(str)
        list of bias files
    flatlist : list(str)
        list of flat field files
    wavelist : list(str)
        list of wavelength calibration files
    orderlist : list(str)
        list of order definition files (for order tracing)
    speclist : list(str)
        list of science files, i.e. observations
    """

    instrument = load_instrument(instrument)
    return instrument.sort_files(input_dir, target, night, mode, **kwargs)


def get_supported_modes(instrument):
    instrument = load_instrument(instrument)
    return instrument.get_supported_modes()


def modeinfo(header, instrument, mode, **kwargs):
    """Add instrument specific information to a header/dict

    Parameters
    ----------
    header : fits.header, dict
        header to add information to
    instrument : str
        instrument name
    mode : str
        instrument mode (e.g. red/blue for HARPS)

    Returns
    -------
    header
        header with added information
    """

    instrument = load_instrument(instrument)
    header = instrument.add_header_info(header, mode, **kwargs)
    return header


def get_wavecal_filename(header, instrument, mode, **kwargs):
    """Get the filename of the pre-existing wavelength solution for the current settings

    Parameters
    ----------
    header : fits.header, dict
        header of the wavelength calibration file
    instrument : str
        instrument name
    mode : str
        instrument mode (e.g. red/blue for HARPS)

    Returns
    -------
    filename : str
        wavelength solution file
    """

    instrument = load_instrument(instrument)
    return instrument.get_wavecal_filename(header, mode, **kwargs)



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
from tqdm import tqdm

from . import util
from .extract import correct_for_curvature, fix_parameters


def rectify_image(
    img, orders, column_range, extraction_width, order_range, tilt=None, shear=None
):
    nord, _ = orders.shape
    nrow, ncol = img.shape
    x = np.arange(ncol)

    extraction_width, column_range, orders = fix_parameters(
        extraction_width, column_range, orders, nrow, ncol, nord
    )

    nord = order_range[1] - order_range[0]
    orders = orders[order_range[0] : order_range[1]]
    column_range = column_range[order_range[0] : order_range[1]]
    extraction_width = extraction_width[order_range[0] : order_range[1]]

    images = {}
    for i in tqdm(range(nord), desc="Order"):
        x_left_lim = column_range[i, 0]
        x_right_lim = column_range[i, 1]

        # Rectify the image, i.e. remove the shape of the order
        # Then the center of the order is within one pixel variations
        ycen = np.polyval(orders[i], x).astype(int)
        yb, yt = ycen - extraction_width[i, 0], ycen + extraction_width[i, 1]
        height = extraction_width[i, 0] + extraction_width[i, 1] + 1
        index = util.make_index(yb, yt, x_left_lim, x_right_lim)
        img_order = img[index]

        # Correct for tilt and shear
        # For each row of the rectified order, interpolate onto the shifted row
        # Masked pixels are set to 0, similar to the summation
        if tilt is not None and shear is not None:
            img_order = correct_for_curvature(
                img_order,
                tilt[i, x_left_lim:x_right_lim],
                shear[i, x_left_lim:x_right_lim],
                extraction_width[i],
            )
        images[i] = img_order

    return images, column_range, extraction_width


def merge_images(images, wave, column_range, extraction_width):
    x_total = sum(img.shape[1] for img in images.values())
    y_max = max(*[img.shape[0] for img in images.values()])
    y_mid = y_max // 2

    combined_img = np.zeros((y_max, x_total))
    wavelength = np.zeros(x_total)

    idx = 0
    x_low = 0
    for iord0, iord1 in zip(range(len(wave) - 1), range(1, len(wave))):

        img0 = images[iord0]
        img1 = images[iord1]

        xwd0, xwd1 = extraction_width[iord0], extraction_width[iord1]
        y0_low = y_mid - xwd0[0]
        y0_high = y_mid + xwd0[1] + 1

        y1_low = y_mid - xwd1[0]
        y1_high = y_mid + xwd1[1] + 1

        # Calculate Overlap
        cr0, cr1 = column_range[iord0], column_range[iord1]
        w0 = wave[iord0][cr0[0] : cr0[1]]
        w1 = wave[iord1][cr1[0] : cr1[1]]

        i0 = np.ma.where((w0 >= np.ma.min(w1)) & (w0 <= np.ma.max(w1)))
        i1 = np.ma.where((w1 >= np.ma.min(w0)) & (w1 <= np.ma.max(w0)))

        if i0[0].size > 0 and i1[0].size > 0:
            # The non overlapping part is just the image
            x_high = i0[0].min()
            combined_img[y0_low:y0_high, idx : idx + x_high - x_low] = img0[
                :, x_low:x_high
            ]
            wavelength[idx : idx + x_high - x_low] = w0[x_low:x_high]

            # for the overlap region use a common wavelength grid
            n_points = (len(i0[0]) + len(i1[0])) // 2
            w_common = np.geomspace(w0[i0][0], w1[i1][-1], num=n_points)

            img0_common = interp1d(
                w0[i0], img0[:, i0[0]], kind="linear", fill_value="extrapolate"
            )(w_common)
            img1_common = interp1d(
                w1[i1], img1[:, i1[0]], kind="linear", fill_value="extrapolate"
            )(w_common)

            # And then simply take the average between the two
            counter_common = np.zeros((y_max, n_points), dtype=int)
            img_common = np.zeros((y_max, n_points))
            img_common[y0_low:y0_high] += img0_common
            counter_common[y0_low:y0_high] += 1
            img_common[y1_low:y1_high] += img1_common
            counter_common[y1_low:y1_high] += 1
            counter_common[counter_common == 0] = 1
            img_common /= counter_common

            combined_img[:, idx + x_low : idx + x_low + n_points] = img_common
            wavelength[idx + x_low : idx + x_low + n_points] = w_common

            idx += x_low + n_points
            x_low = i1[0].max()
        else:
            x_high = img0.shape[1]
            combined_img[y0_low:y0_high, idx : idx + x_high] = img0
            wavelength[idx : idx + x_high] = w0
            idx += x_high
            x_low = 0

    img0 = images[len(wave) - 1]
    y0 = img0.shape[0]
    y0_low = (y_max - y0) // 2
    y0_high = y0 + y0_low
    cr0 = column_range[iord0]
    w0 = wave[iord0][cr0[0] : cr0[1]]

    x_high = img0.shape[1]
    combined_img[y0_low:y0_high, idx : idx + x_high - x_low] = img0[:, x_low:x_high]
    wavelength[idx : idx + x_high - x_low] = w0[x_low:x_high]

    idx += x_high - x_low
    combined_img = combined_img[:, :idx]
    wavelength = wavelength[:idx]

    return wavelength, combined_img



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Builds the C library that contains the extraction algorithm

This module prepares and builds the C libary containing the curved
(and vertical) extraction algorithm using CFFI.
It also prepares the ffibuilder objects for setup.py,
so that the library is compiled on installation.


The user can also call the Module as a script to compile the
C libraries again.

Attributes
----------
ffi_builder_vertical : FFI
    CFFI Builder for the vertical extraction algorithm
ffi_builder_curved : FFI
    CFFI Builder for the curved extraction algorithm
"""

import logging
import os

from cffi import FFI

logger = logging.getLogger(__name__)


CWD = os.path.dirname(__file__)
CWD = os.path.abspath(CWD)
release_path = os.path.join(CWD, "Release")

print("Include dir: ", CWD)
print("Release dir: ", release_path)


ffibuilder_vertical = FFI()
with open(os.path.join(CWD, "slit_func_bd.h")) as f:
    ffibuilder_vertical.cdef(f.read())
with open(os.path.join(CWD, "slit_func_bd.c"), "r") as f:
    ffibuilder_vertical.set_source(
        "_slitfunc_bd",
        f.read(),
        include_dirs=[CWD, release_path],
        depends=["slit_func_bd.h"],
    )

ffibuilder_curved = FFI()
with open(os.path.join(CWD, "slit_func_2d_xi_zeta_bd.h")) as f:
    ffibuilder_curved.cdef(f.read())
with open(os.path.join(CWD, "slit_func_2d_xi_zeta_bd.c"), "r") as f:
    ffibuilder_curved.set_source(
        "_slitfunc_2d",
        f.read(),
        include_dirs=[CWD, release_path],
        depends=["slit_func_2d_xi_zeta_bd.h"],
    )


def build():
    """Builds the C slitfunc library"""
    logger.info("Building required C libraries, this might take a few seconds")

    old_cwd = os.getcwd()
    path = os.path.abspath(CWD)
    os.chdir(path)

    ffibuilder_vertical.compile(verbose=True, debug=False)
    ffibuilder_curved.compile(verbose=True, debug=False)

    os.chdir(old_cwd)


if __name__ == "__main__":  # pragma: no cover
    build()



# -*- coding: utf-8 -*-
"""
Contains functions to read and modify echelle structures, just as in reduce

Mostly for compatibility reasons
"""

import logging

import astropy.io.fits as fits
import numpy as np
import scipy.constants

logger = logging.getLogger(__name__)


class Echelle:
    def __init__(self, head={}, filename="", data={}):
        self.filename = filename
        self.header = head
        self._data = data

    @property
    def nord(self):
        if "spec" in self._data.keys():
            return self._data["spec"].shape[0]
        return None

    @property
    def ncol(self):
        if "spec" in self._data.keys():
            return self._data["spec"].shape[1]
        return None

    @property
    def spec(self):
        if "spec" in self._data.keys():
            return self._data["spec"]
        else:
            return None

    @spec.setter
    def spec(self, value):
        self._data["spec"] = value

    @property
    def sig(self):
        if "sig" in self._data.keys():
            return self._data["sig"]
        else:
            return None

    @sig.setter
    def sig(self, value):
        self._data["sig"] = value

    @property
    def wave(self):
        if "wave" in self._data.keys():
            return self._data["wave"]
        else:
            return None

    @wave.setter
    def wave(self, value):
        self._data["wave"] = value

    @property
    def cont(self):
        if "cont" in self._data.keys():
            return self._data["cont"]
        else:
            return None

    @cont.setter
    def cont(self, value):
        self._data["cont"] = value

    @property
    def columns(self):
        if "columns" in self._data.keys():
            return self._data["columns"]
        else:
            return None

    @columns.setter
    def columns(self, value):
        self._data["columns"] = value

    @property
    def mask(self):
        if "mask" in self._data.keys():
            return self._data["mask"]
        else:
            return None

    @mask.setter
    def mask(self, value):
        self._data["mask"] = value

    def __getitem__(self, index):
        return self._data[index]

    def __setitem__(self, index, value):
        self._data[index] = value

    def __delitem__(self, index):
        del self._data[index]

    def __contains__(self, index):
        return index in self._data.keys()

    @staticmethod
    def read(
        fname,
        extension=1,
        raw=False,
        continuum_normalization=True,
        barycentric_correction=True,
        radial_velociy_correction=True,
    ):
        """
        Read data from an echelle file
        Expand wavelength and continuum polynomials
        Apply barycentric/radial velocity correction
        Apply continuum normalization

        Will load any fields in the binary table, however special attention is given only to specific names:
        "SPEC"    : Spectrum
        "SIG"     : Sigma, i.e. (absolute) uncertainty
        "CONT"    : Continuum
        "WAVE"    : Wavelength solution
        "COLUMNS" : Column range

        Parameters
        ----------
        fname : str
            filename to load
        extension : int, optional
            fits extension of the data within the file (default: 1)
        raw : bool, optional
            if true apply no corrections to the data (default: False)
        continuum_normalization : bool, optional
            apply continuum normalization (default: True)
        barycentric_correction : bool, optional
            apply barycentric correction (default: True)
        radial_velociy_correction : bool, optional
            apply radial velocity correction (default: True)

        Returns
        -------
        ech : obj
            Echelle structure, with data contained in attributes
        """

        hdu = fits.open(fname)
        header = hdu[0].header
        data = hdu[extension].data

        _data = {column.lower(): data[column][0] for column in data.dtype.names}
        ech = Echelle(filename=fname, head=header, data=_data)
        nord, ncol = ech.nord, ech.ncol

        if not raw:
            if "spec" in ech:
                base_order = header.get("obase", 1)
                ech["orders"] = np.arange(base_order, base_order + nord)

            # Wavelength
            if "wave" in ech:
                ech["wave"] = expand_polynomial(ncol, ech["wave"])

                # Correct for radial velocity and barycentric correction
                # + : away from observer
                # - : towards observer
                velocity_correction = 0
                if barycentric_correction:
                    velocity_correction -= header.get("barycorr", 0)
                    header["barycorr"] = 0
                if radial_velociy_correction:
                    velocity_correction += header.get("radvel", 0)
                    header["radvel"] = 0

                speed_of_light = scipy.constants.speed_of_light * 1e-3
                ech["wave"] *= 1 + velocity_correction / speed_of_light

            # Continuum
            if "cont" in ech:
                ech["cont"] = expand_polynomial(ncol, ech["cont"])

            # Create Mask, based on column range
            if "columns" in ech:
                ech["mask"] = np.full((nord, ncol), True)
                for iord in range(nord):
                    ech["mask"][
                        iord, ech["columns"][iord, 0] : ech["columns"][iord, 1]
                    ] = False

                if "spec" in ech:
                    ech["spec"] = np.ma.masked_array(ech["spec"], mask=ech["mask"])
                if "sig" in ech:
                    ech["sig"] = np.ma.masked_array(ech["sig"], mask=ech["mask"])
                if "cont" in ech:
                    ech["cont"] = np.ma.masked_array(ech["cont"], mask=ech["mask"])
                if "wave" in ech:
                    ech["wave"] = np.ma.masked_array(ech["wave"], mask=ech["mask"])

            # Apply continuum normalization
            if continuum_normalization and "cont" in ech:
                if "spec" in ech:
                    ech["spec"] /= ech["cont"]
                if "sig" in ech:
                    ech["sig"] /= ech["cont"]

        return ech

    def save(self, fname):
        """
        Save data in an Echelle fits, i.e. a fits file with a Binary Table in Extension 1

        Parameters
        ----------
        fname : str
            filename
        """
        save(fname, self.header, **self._data)


def calc_2dpolynomial(solution2d):
    """Expand a 2d polynomial, where the data is given in a REDUCE make_wave format
    Note that the coefficients are for order/100 and column/1000 respectively, where the order is counted from order base up

    Parameters
    ----------
    solution2d : array
        data in a REDUCE make_wave format:
        0: version
        1: number of columns
        2: number of orders
        3: order base, i.e. 0th order number (should not be 0)
        4-6: empty
        7: number of cross coefficients
        8: number of column only coefficients
        9: number of order only coefficients
        10: coefficient - constant
        11-x: column coefficients
        x-y : order coefficients
        z-  : cross coefficients (xy, xy2, x2y, x2y2, xy3, x3y), with x = orders, y = columns

    Returns
    -------
    poly : array[nord, ncol]
        expanded polynomial
    """

    # make wave style 2d fit
    ncol = int(solution2d[1])
    nord = int(solution2d[2])
    order_base = int(solution2d[3])
    deg_cross, deg_column, deg_order = (
        int(solution2d[7]),
        int(solution2d[8]),
        int(solution2d[9]),
    )
    coeff_in = solution2d[10:]

    coeff = np.zeros((deg_order + 1, deg_column + 1))
    coeff[0, 0] = coeff_in[0]
    coeff[0, 1:] = coeff_in[1 : 1 + deg_column]
    coeff[1:, 0] = coeff_in[1 + deg_column : 1 + deg_column + deg_order]
    if deg_cross in [4, 6]:
        coeff[1, 1] = coeff_in[deg_column + deg_order + 1]
        coeff[1, 2] = coeff_in[deg_column + deg_order + 2]
        coeff[2, 1] = coeff_in[deg_column + deg_order + 3]
        coeff[2, 2] = coeff_in[deg_column + deg_order + 4]
    if deg_cross == 6:
        coeff[1, 3] = coeff_in[deg_column + deg_order + 5]
        coeff[3, 1] = coeff_in[deg_column + deg_order + 6]

    x = np.arange(order_base, order_base + nord, dtype=float)
    y = np.arange(ncol, dtype=float)

    poly = np.polynomial.polynomial.polygrid2d(x / 100, y / 1000, coeff) / x[:, None]

    return poly


def calc_1dpolynomials(ncol, poly):
    """Expand a set of 1d polynomials (one per order) seperately

    Parameters
    ----------
    ncol : int
        number of columns
    poly : array[nord, degree]
        polynomial coefficients

    Returns
    -------
    poly : array[nord, ncol]
        expanded polynomials
    """

    nord = poly.shape[0]
    x = np.arange(ncol)
    result = np.zeros((nord, ncol))
    for i, coef in enumerate(poly):
        result[i] = np.polyval(coef, x)
    return result


def expand_polynomial(ncol, poly):
    """Checks if and how to expand data poly, then expands the data if necessary

    Parameters
    ----------
    ncol : int
        number of columns in the image
    poly : array[nord, ...]
        polynomial coefficients to expand, or already expanded data

    Returns
    -------
    poly : array[nord, ncol]
        expanded data
    """

    if poly.ndim == 1:
        poly = calc_2dpolynomial(poly)
    elif poly.shape[1] < 20:
        poly = calc_1dpolynomials(ncol, poly)
    return poly


def read(fname, **kwargs):
    return Echelle.read(fname, **kwargs)


def save(fname, header, **kwargs):
    """Save data in an Echelle fits, i.e. a fits file with a Binary Table in Extension 1

    The data is passed in kwargs, with the name of the binary table column as the key
    Floating point data is saved as float32 (E), Integer data as int16 (I)

    Parameters
    ----------
    fname : str
        filename
    header : fits.header
        FITS header
    **kwargs : array[]
        data to be saved in the file
    """

    if not isinstance(header, fits.Header):
        header = fits.Header(cards=header)

    primary = fits.PrimaryHDU(header=header)

    columns = []
    for key, value in kwargs.items():
        if value is None:
            continue
        arr = value.ravel()[None, :]

        if np.issubdtype(arr.dtype, np.floating):
            arr = arr.astype(np.float32)
            dtype = "E"
        elif np.issubdtype(arr.dtype, np.integer):
            arr = arr.astype(np.int16)
            dtype = "I"
        elif np.issubdtype(arr.dtype, np.bool_):
            arr = arr.astype(np.bool_)
            dtype = "B"
        else:
            raise TypeError(f"Could not understand dtype {arr.dtype}")

        form = "%i%s" % (value.size, dtype)
        dim = str(value.shape[::-1])
        columns += [fits.Column(name=key.upper(), array=arr, format=form, dim=dim)]

    table = fits.BinTableHDU.from_columns(columns)

    hdulist = fits.HDUList(hdus=[primary, table])
    hdulist.writeto(fname, overwrite=True, output_verify="silentfix+ignore")



# -*- coding: utf-8 -*-

import astropy.io.fits as fits
import numpy as np
import pytest

from pyreduce import combine_frames


@pytest.fixture
def buffer():
    return np.array([np.arange(10), np.arange(1, 11), np.arange(2, 12)])


@pytest.fixture
def size():
    return 3


def create_file(file, nx=100, ny=100, ovscx=5):
    img = np.full((ny, nx), 10)
    data = img + np.random.randint(0, 20, size=img.shape)
    head = fits.Header(
        cards={
            "ESO DET OUT1 PRSCX": 0,
            "ESO DET OUT1 OVSCX": ovscx,
            "ESO DET OUT1 CONAD": 1,
            "ESO DET OUT1 RON": 0,
            "EXPTIME": 1,
            "RA": 100,
            "DEC": 51,
            "DATE-OBS": "2011-07-17T21:09:37.7545",
        }
    )
    fits.writeto(file, data=data, header=head)
    return data, head


def test_running_median(buffer, size):
    result = combine_frames.running_median(buffer, size)
    compare = np.array([np.arange(1, 9), np.arange(2, 10), np.arange(3, 11)])

    assert np.array_equal(result, compare)


def test_running_mean(buffer, size):
    result = combine_frames.running_median(buffer, size)
    compare = np.array([np.arange(1, 9), np.arange(2, 10), np.arange(3, 11)])

    assert np.array_equal(result, compare)


def test_calculate_probability(buffer):
    window = 1
    compare = np.array(
        [np.arange(1, 9), np.arange(2, 10), np.arange(3, 11)], dtype=float
    )

    result = combine_frames.calculate_probability(buffer, window, "sum")
    weights = np.sum(compare, axis=0)
    assert np.allclose(result, compare / weights)

    result = combine_frames.calculate_probability(buffer, window, "median")
    weights = np.mean(compare, axis=0)
    assert np.allclose(result, compare / weights)


def test_combine_frames(tempfiles):
    for f in tempfiles:
        create_file(f, 100, 100, 5)

    combine, chead = combine_frames.combine_frames(
        tempfiles, "UVES", "middle", 0, window=5
    )

    assert combine.shape[0] == 100 - 5
    assert combine.shape[1] == 100

    assert chead["exptime"] == len(tempfiles)


def test_nofiles():
    files = []
    with pytest.raises(ValueError):
        combine_frames.combine_frames(files, "UVES", "middle", 0, window=5)


def test_onefile(tempfiles):
    tempfiles = tempfiles[:1]
    nx, ny = 110, 100
    compare, head = create_file(tempfiles[0], nx, ny, 5)
    compare = np.rot90(compare, -1)[:-5]

    combine, chead = combine_frames.combine_frames(
        tempfiles, "UVES", "middle", 0, window=5
    )

    assert combine.shape[0] == nx - 5
    assert combine.shape[1] == ny
    assert np.allclose(combine, compare)

    for key, value in head.items():
        assert chead[key] == value


def test_twofiles(tempfiles):
    tempfiles = tempfiles[:2]

    for f in tempfiles:
        create_file(f, 100, 100, 5)

    combine, chead = combine_frames.combine_frames(
        tempfiles, "UVES", "middle", 0, window=5
    )

    assert combine.shape[0] == 100 - 5
    assert combine.shape[1] == 100

    assert chead["exptime"] == len(tempfiles)


def test_bad_window_size(tempfiles):
    for f in tempfiles:
        create_file(f, 100, 100, 5)

    combine, chead = combine_frames.combine_frames(
        tempfiles, "UVES", "middle", 0, window=80
    )
    assert combine.shape[0] == 100 - 5
    assert combine.shape[1] == 100

    assert chead["exptime"] == len(tempfiles)


def test_normal_orientation(tempfiles):
    for f in tempfiles:
        create_file(f, 100, 100, 0)

    combine, chead = combine_frames.combine_frames(
        tempfiles, "CRIRES_PLUS", "J1228_OPEN_det1", 0, window=10
    )
    assert combine.shape[0] == 100 - 10  # there is a 5 pixel cutoff on each side
    assert combine.shape[1] == 100 - 10  # there is a 5 pixel cutoff on each side

    assert chead["exptime"] == len(tempfiles)
